''*******************************************************************
''*       Title: VGA display 40x15 (single cog, ROM font, palette)  *
''*    FileName: VGA_40x15_driver.spin                              *
''* Description: Video driver and pixel generator                   *
''*      Author: Marko Lukat (c) 2013                               *
''*******************************************************************
'' Revision History
''-----------------
''   Release: 2013/10/01 - Version: 0.1.pal.1 - by Marko Lukat 
''     (original) initial version (640x480@60Hz timing, %11 sync locked)
''   Release: 2013/10/04 - Version: 0.2.pal.2    Marko Lukat 
''        (1) change: now uses palette (4/4 split)
''   Release: 2013/10/05 - Version: 0.6.pal.5 - by Marko Lukat 
''        (2) change: unknown
''   Release: 2013/10/11 - Version: 0.7.pal.5 - by PropGuy2 
''        (3) changed filename to: VGA_40x15_driver.spin
''        (4) VGA_40x15_driver: moved CON(zero..alias) to beginning of listing
''-----------------
'' long[par][0]:  screen: [!Z]:addr = 16:16 -> zero (accepted), 2n
'' long[par][1]: palette: [!Z]:addr = 16:16 -> zero (accepted), 2n
'' long[par][2]:  colour: [!Z]:addr = 16:16 -> zero (accepted), 2n
'' long[par][3]: frame indicator
''-----------------
''
CON
  zero    = $1F0                      'par (dst only)
  vpin    = $0FC                      'pin group mask
  vgrp    = 2                         'pin group
  sgrp    = 2                         'pin group sync
  hv_idle = $01010101 * %11 {%hv}     'h/v sync inactive
  dcolour = %%0220_0010               'default colour
  
  res_x   = 640                       'screen width
  res_y   = 480                       'screen height
  res_m   = 4                         'colour bytes          
                                                                
  alias   = 0                         'alias

CON
  ID_0 = $30B3309C
  ID_1 = $9ED2732B
  ID_2 = $38343032                    'cog binary magic number (2048)

  OVERLAY = %00000000_00000001        'cog binary is an overlay
  MAPPING = %00000000_00000010        'translation table is present

OBJ
'' This is not a top level object.

PUB null
'' This is not a top level object.
  
PUB init(ID, mailbox)                 'initialize driver
  return launch(ID, @driver, mailbox)

PUB launch(ID, code, data)            'launch VGA driver cog
'' PASM quick launch using a specific or the next available ID.
''
'' parameters
''       ID: cog ID
''           0..7: coginit, otherwise cognew (may fail)
''     code: address of code fragment (4n)
''     data: cognew/coginit parameter (4n)
''
'' result
''     == 0: [ABORT] thread creation failed (cognew only)
''     <> 0: thread/cog ID + 1

  ifnot (ID >> 3)
    coginit(ID++, code, data)
  elseifnot ID := cognew(code, data) + 1
    abort
  return ID
  
DAT             org     0                       ' cog binary header

header_2048     long    ID_2                    ' magic number for a cog binary
                word    header_size             ' header size
                word    MAPPING                 ' flags      >>> was:  "system#"
                word    0, 0                    ' start register, register count

                word    @__table - @header_2048 ' translation table byte offset

header_size     fit     16
                
DAT             org     0                       ' video driver

driver          jmpret  $, #setup               '  -4   once

                mov     dira, mask              ' drive outputs

' horizontal timing 640(640)  1(16) 6(96)  3(48)
'   vertical timing 480(480) 10(10) 2(2)  33(33)

vsync           mov     ecnt, #10
                call    #blank                  ' front porch
                djnz    ecnt, #$-1

                xor     sync, #$0101            ' active
                                                '                       mov     ecnt, #2  
                call    #blank                  ' |                     call    #blank    
                call    #blank                  ' vertical sync         djnz    ecnt, #$-1

                xor     sync, #$0101            ' inactive

' Start fetch during last 32 blank lines.

                call    #blank                  ' back porch (1st line)

                mov     eins, scrn              ' screen base address
                mov     vier, indx              ' colour buffer
                mov     drei, #res_y            ' max visible scanlines

                mov     scnt, #33 -1            ' scnt controls fetch (32..1)
                call    #blank                  ' back porch
                djnz    scnt, #$-1

' Vertical sync chain done, do visible area.

:line           mov     frqb, #0 wc             ' initial font row index
                mov     scnt, #32               ' font size
                max     scnt, drei              ' limit against what's left
                sub     drei, scnt              ' update what's left

:scan           waitcnt cnt, #0                 ' re-sync after back porch              (##)

                mov     outa, idle              ' take over sync lines
                mov     vcfg, vcfg_norm         ' disconnect from video h/w             (&&)

                call    #emit                   ' |
                call    #hsync                  ' display scanline
                
                add     frqb, #2                ' next row in character definition(s)
                djnz    scnt, #:scan            ' repeat for font size
                tjnz    drei, #:line            ' repeat for all rows

                wrlong  cnt, fcnt_              ' announce vertical blank

                rdlong  plte, plte_ wz          ' fetch pending updates
        if_nz   wrlong  zero, plte_             ' acknowledge user palette
        
                jmp     #vsync                  ' next frame


blank           mov     vscl, line              ' 256/640
                waitvid sync, #%%000            ' latch blank line

                call    #palette                ' override current palette
                
hsync           mov     vscl, wrap              ' horizontal sync
                waitvid sync, wrap_value

                mov     vcfg, vcfg_sync         ' drive sync lines                      (&&)
                mov     outa, #0                ' stop interfering

                mov     cnt, cnt                ' record sync point                     (##)
                add     cnt, #9{14}+412           

                call    #fetch
hsync_ret
blank_ret       ret


emit            mov     vscl, hvis              ' 1/16
                mov     ecnt, #40               ' character count

        if_nc   xor     $+2, #addr^(addr+40)    ' address array |
        if_nc   xor     $+2, #cols^(cols+40)    ' colour array  | primary <> secondary

                movs    :one, #addr             ' |
                movd    :two, #cols             ' reset
                
:one            mov     phsb, addr              ' character pixmap address
                rdlong  temp, phsb              ' read from address + 2frqb (row adjustment)
:two            waitvid cols, temp                

                add     :one, #1                ' next address |
                add     :two, dst1              ' next colour  | cog array

                djnz    ecnt, #:one             ' next 16px
    
emit_ret        jmpret  zero, #0-0 wc,nr        ' set carry


fetch           cmp     drei, #0 wz             ' enabled?
        if_e    jmp     fetch_ret

                cmp     scnt, #21 wz,wc         ' 21..32
        if_ae   jmp     #f_21                   ' re-initialise sequence

                call    #ascii
                call    #ascii
                
fetch_ret       jmpret  zero, #0-0 wc,nr        ' set carry


f_21    if_e    xor     $+2, #addr^(addr+40)    ' |
        if_e    xor     $+2, #cols^(cols+40)    ' swap buffers

        if_e    movd    f_00, #addr             ' |
        if_e    movd    f_01, #cols             ' preset filler code
                jmpret  zero, fetch_ret wc,nr   ' set carry


ascii           rdbyte  temp, vier              ' fetch colour index
                ror     temp, #4                ' extract foreground
                add     temp, #ntry             ' apply table offset

                movs    f_fg, temp              ' inject address
                shr     temp, #28               ' extract background
                add     temp, #ntry +16         ' apply table offset
                movs    f_bg, temp              ' inject address
                
                rdbyte  temp, eins              ' fetch character
                add     eins, #1                ' advance address

                shr     temp, #1 wc             ' even/odd
                or      temp, #%1_00000000      ' ROM base address
                shl     temp, #7                ' final location

f_00            mov     0-0, temp
                add     $-1, dst1

f_fg            mov     zwei, 0-0               ' read foreground
f_bg            or      zwei, 0-0               ' read background
                add     vier, #1                ' advance address

                mov     temp, zwei              ' $0000BBAA     
        if_c    shr     temp, #8
                shl     temp, #16               ' $BBAA0000/$00BB0000
                or      temp, zwei              ' $BBAABBAA/$00BBBBAA

        if_c    shl     temp, #8                ' $BBBBAA00
        if_c    and     zwei, #$FF
        if_c    or      temp, zwei              ' $BBBBAAAA
                
f_01            mov     0-0, temp
                add     $-1, dst1
                
ascii_ret       ret


palette         cmp     plte, #0 wz             ' pending?
        if_z    jmp     palette_ret             ' ... no

                rdbyte  ntry+$00, plte
                shl     ntry+$00, #8
                add     plte, #1
                rdbyte  ntry+$01, plte
                shl     ntry+$01, #8
                add     plte, #1
                rdbyte  ntry+$02, plte
                shl     ntry+$02, #8
                add     plte, #1
                rdbyte  ntry+$03, plte
                shl     ntry+$03, #8
                add     plte, #1
                rdbyte  ntry+$04, plte
                shl     ntry+$04, #8
                add     plte, #1
                rdbyte  ntry+$05, plte
                shl     ntry+$05, #8
                add     plte, #1
                rdbyte  ntry+$06, plte
                shl     ntry+$06, #8
                add     plte, #1
                rdbyte  ntry+$07, plte
                shl     ntry+$07, #8
                add     plte, #1
                rdbyte  ntry+$08, plte
                shl     ntry+$08, #8
                add     plte, #1
                rdbyte  ntry+$09, plte
                shl     ntry+$09, #8
                add     plte, #1
                rdbyte  ntry+$0A, plte
                shl     ntry+$0A, #8
                add     plte, #1
                rdbyte  ntry+$0B, plte
                shl     ntry+$0B, #8
                add     plte, #1
                rdbyte  ntry+$0C, plte
                shl     ntry+$0C, #8
                add     plte, #1
                rdbyte  ntry+$0D, plte
                shl     ntry+$0D, #8
                add     plte, #1
                rdbyte  ntry+$0E, plte
                shl     ntry+$0E, #8
                add     plte, #1
                rdbyte  ntry+$0F, plte
                shl     ntry+$0F, #8
                add     plte, #1

                rdbyte  ntry+$10, plte
                add     plte, #1
                rdbyte  ntry+$11, plte
                add     plte, #1
                rdbyte  ntry+$12, plte
                add     plte, #1
                rdbyte  ntry+$13, plte
                add     plte, #1
                rdbyte  ntry+$14, plte
                add     plte, #1
                rdbyte  ntry+$15, plte
                add     plte, #1
                rdbyte  ntry+$16, plte
                add     plte, #1
                rdbyte  ntry+$17, plte
                add     plte, #1
                rdbyte  ntry+$18, plte
                add     plte, #1
                rdbyte  ntry+$19, plte
                add     plte, #1
                rdbyte  ntry+$1A, plte
                add     plte, #1
                rdbyte  ntry+$1B, plte
                add     plte, #1
                rdbyte  ntry+$1C, plte
                add     plte, #1
                rdbyte  ntry+$1D, plte
                add     plte, #1
                rdbyte  ntry+$1E, plte
                add     plte, #1
                rdbyte  ntry+$1F, plte

                mov     plte, #0                ' done
palette_ret     ret

' initialised data and/or presets

idle            long    hv_idle
sync            long    hv_idle ^ $0200

frqx            long    $1423D70A               ' 25.175MHz
                        
wrap_value      long    %%0001111110            ' horizontal sync pulse (1/6/3 reverse)
wrap            long    16 << 12 | 160          '  16/160
hvis            long     1 << 12 | 16           '   1/16
line            long     0 << 12 | 640          ' 256/640

vcfg_norm       long    %0_01_1_00_000 << 23 | vgrp << 9 | vpin
vcfg_sync       long    %0_01_1_00_000 << 23 | sgrp << 9 | %11

mask            long    vpin << (vgrp * 8) | %11 << (sgrp * 8)

scrn_           long    +0                      ' |
plte_           long    +4                      ' |
indx_           long    +8                      ' |
fcnt_           long    12                      ' mailbox addresses (local copy)

drei            long    0                       ' no fetch on startup
dst1            long    1 << 9                  ' dst +/-= 1

ntry            long    (dcolour & $FF00)[16]   ' foreground
                long    (dcolour & $00FF)[16]   ' background
                
' Stuff below is re-purposed for temporary storage.

setup           add     scrn_, par              ' @long[par][0]
                add     plte_, par              ' @long[par][1]
                add     indx_, par              ' @long[par][2]
                add     fcnt_, par              ' @long[par][3]

                rdlong  scrn, scrn_ wz          ' get screen address                    (%%)
        if_nz   wrlong  zero, scrn_             ' acknowledge screen buffer setup

                rdlong  indx, indx_ wz          ' get font colour                       (%%)
        if_nz   wrlong  zero, indx_             ' acknowledge font colour

                rdlong  plte, plte_ wz          ' get user palette                      (%%)
        if_nz   wrlong  zero, plte_             ' acknowledge user palette
        if_nz   call    #palette                ' initial override
        
' Upset video h/w and relatives.

                movi    ctrb, #%0_11111_000     ' LOGIC always (loader support)
                movi    ctra, #%0_00001_101     ' PLL, VCO/4
                mov     frqa, frqx              ' 25.175MHz
                
                mov     vscl, hvis              ' 1/16
                mov     vcfg, vcfg_sync         ' VGA, 4 colour mode

' Setup complete, do the heavy lifting upstairs ...

                jmp     %%0                     ' return

                fit
                
' uninitialised data and/or temporaries

                org     setup

scrn            res     1                       ' screen buffer         < setup +4      (%%)
indx            res     1                       ' colour buffer         < setup +6      (%%)
plte            res     1                       ' user palette          < setup +8      (%%)
ecnt            res     1                       ' element count
scnt            res     1                       ' scanlines (per char)
temp            res     1

eins            res     1
zwei            res     1
vier            res     1

addr            res     40 *2
cols            res     40 *2

tail            fit
                
DAT                                             ' translation table

__table         word    (@__names - @__table)/2

                word    res_x
                word    res_y
                word    res_m
                
__names         byte    "res_x", 0
                byte    "res_y", 0
                byte    "res_m", 0

DAT